/** @file
 *  @brief Bluetooth Mesh Porting APIs.
 */

/*
 * Copyright (c) 2017 Intel Corporation
 *
 * SPDX-License-Identifier: Apache-2.0
 */
#ifndef __BT_MESH_PORTING_H
#define __BT_MESH_PORTING_H

#ifdef __cplusplus
extern "C" {
#endif

void mesh_adv_thread(void *args);

#ifdef __cplusplus
}
#endif

/**
 * @}
 */

#endif /* __BT_MESH_PORTING_H */
